#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['physical_education', 'physical_education.models']

package_data = \
{'': ['*']}

install_requires = \
['dill ~= 0.3',
 'matplotlib ~= 3.1',
 'numpy ~= 1.10',
 'pyomo == 5.6.8',
 'sympy ~= 1.3',
 'typing_extensions ~= 3.0']

entry_points = \
{'console_scripts': ['flit = flit:main']}

setup(name='physical_education',
      version='0.1.1',
      description='A library for research into legged critters',
      author='Alexander Knemeyer',
      author_email='alknemeyer@hey.com',
      url='https://github.com/alknemeyer/physical_education',
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      entry_points=entry_points,
      python_requires='>=3.6',
     )
